#undef NDEBUG
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "bbqueue.h"

// ns2 requests its own random function
#ifdef HAVE_OTCL_H
#include "random.h"
#endif

#ifndef panic
#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); exit(1); } while (0)
#endif

BroadbandQueue::BroadbandQueue()
{
  delayMicros = 0;
  capacityPackets = -1;
  capacityBytes = 10*1500;
  lineRateBps = 4000000;
  reset();
}

void BroadbandQueue::reset()
{
  lastArrival = 0;
  numEntries = 0;
  currentBurstRequestTime = -1;
  currentBurstXmitTime = -1;
  currentBurstSizeBytes = 0;
  currentBurstID = 0;
  nextMediumFree = 0;
}

void BroadbandQueue::getQueueSizeAt(long long now, int *packets, int *bytes)
{
  *packets = 0;
  *bytes = 0;

  for (int i=0; i<numEntries; i++) {
    if ((queue[i].txmit >= now) && !queue[i].inFlight) {
      *bytes += queue[i].sizeBytes;
      (*packets) ++;
    }
  }
}

bool BroadbandQueue::append(int id, int size, long long now)
{

  long long xmitReady = now;
  assert(now >= lastArrival);
  lastArrival = now;
  
  /* Emulate delay */
  
  xmitReady += delayMicros;
  
  /* Emulate queue capacity. This may decide to drop the packet if the queue is full. */
  
  int queueSizeBytes = 0;
  int queueSizePackets = 0;
  getQueueSizeAt(xmitReady, &queueSizePackets, &queueSizeBytes);
  
  if ((capacityBytes >= 0) && ((queueSizeBytes + size) > capacityBytes))
    return false;
  if ((capacityPackets >= 0) && ((queueSizePackets + 1) > capacityPackets))
    return false;

  /* Okay; the packet made it safely into the queue and won't be dropped */

  /* If the packet arrived BEFORE we will request bandwidth for the next burst, we can
     simply make that burst longer and request more bandwidth. */

  if (currentBurstRequestTime >= 0) {
    currentBurstRequestTime = -1;
    currentBurstXmitTime = -1;
    currentBurstSizeBytes = 0;
  }
    
    /* ... and we start a new burst. */
    
  currentBurstID ++;
  currentBurstRequestTime = (xmitReady<nextMediumFree) ? nextMediumFree : xmitReady;
  currentBurstSizeBytes = size;
  currentBurstXmitTime = currentBurstRequestTime;
  
  nextMediumFree = currentBurstXmitTime + (long long)(size * 8000000.0/lineRateBps);
  xmitReady = currentBurstXmitTime;
  

  /* Now we queue the packet along with its transmission time */

  if (numEntries >= MAX_ENTRIES)
    panic("Queue overflow");
    
  queue[numEntries].id = id;
  queue[numEntries].sizeBytes = size;
  queue[numEntries].txmit = xmitReady;
  queue[numEntries].burstID = currentBurstID;
  queue[numEntries].inFlight = false;
  numEntries ++;

  return true;
}

long long BroadbandQueue::nextEvent()
{
  if (!numEntries)
    return -1;
  
  return queue[0].txmit;
}

int BroadbandQueue::getNextPacket(long long now)
{
  if ((currentBurstRequestTime >= 0) && (now >= currentBurstRequestTime)) {
    currentBurstRequestTime = -1;
    currentBurstSizeBytes = 0;
  }

  assert(now >= lastArrival);
  lastArrival = now;
  
  if ((!numEntries) || (queue[0].txmit > now))
    return -1;
    
  if (!queue[0].inFlight) {
    for (int i=0; i<numEntries; i++) {
      if (queue[i].burstID != queue[0].burstID)
        break;
      queue[i].inFlight = true;
    }
  }
    
  int id = queue[0].id;
  for (int i=1; i<numEntries; i++)
    queue[i-1] = queue[i];
  numEntries --;
  
  return id;
}

void BroadbandQueue::dump(FILE *stream) 
{
  fprintf(stream, "=== inQueue=%d curReqTime=%lld curXmitTime=%lld mediaFree=%lld curSize=%d curID=%d\n", 
    numEntries, currentBurstRequestTime, currentBurstXmitTime, nextMediumFree,
    currentBurstSizeBytes, currentBurstID
  );
  
  for (int i=0; i<numEntries; i++)
    fprintf(stream, "#%02d: id=%d size=%d txmit=%lld burstID=%d%s\n", i, queue[i].id, queue[i].sizeBytes, queue[i].txmit, queue[i].burstID, queue[i].inFlight ? " XMIT" : "");
}
