#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "parameters.h"

#ifdef panic
#undef panic
#endif
#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); abort(); } while (0)

ParameterFile::ParameterFile(const char *filename)
{
  numCategories = 0;
  numDefinitions = 0;
  selectedCategory = -1;

  FILE *infile;
  if (filename) {
    strncpy(paramFileName, filename, sizeof(paramFileName));
    infile = fopen(filename, "r");
    if (!infile)
      panic("Cannot read parameter file: '%s'", filename);
  } else {
    strcpy(paramFileName, "stdin");
    infile = stdin;
  }
  
  int lineno = 0;
  char linebuf[500];
  while (fgets(linebuf, sizeof(linebuf), infile)) {
    lineno ++;
    char *p = &linebuf[0];
    while (*p) {
      if (strchr("\r\n#", *p))
        *p = 0;
      else 
        p++;
    }
    
    p = &linebuf[0];
    while (*p && isspace(*p))
      p++;
    if (!*p)
      continue;
    if (*p == '[') {
      if (numCategories >= MAX_CATEGORIES)
        panic("%s:%d: Too many categories (max %d)", filename, lineno, MAX_CATEGORIES);
      char *thisCategoryName = &p[1];
      strtok(thisCategoryName, "]");
      for (int i=0; i<numCategories; i++)
        if (!strcmp(thisCategoryName, categoryName[i]))
          panic("%s:%d: Duplicate category: '%s'", filename, lineno, thisCategoryName);
          
      strcpy(categoryName[numCategories++], thisCategoryName);
    } else {
      if (!strchr(p, '='))
        panic("%s:%d: Unknown setting: '%s'", filename, lineno, p);
      if (numCategories == 0)
        panic("%s:%d: Global settings not allowed", filename, lineno);
      if (numDefinitions >= MAX_DEFINITIONS)
        panic("%s:%d: Too many definitions (max %d)", filename, lineno, MAX_DEFINITIONS);
      
      char *equ = strchr(p, '=');
      *equ++ = 0;
      char *namepart = p;
      char *valuepart = equ;
      while (*namepart && isspace(*namepart))
        namepart ++;
      while (*valuepart && isspace(*valuepart))
        valuepart ++;
      while (*namepart && isspace(namepart[strlen(namepart)-1]))
        namepart[strlen(namepart)-1] = 0;
      while (*valuepart && isspace(valuepart[strlen(valuepart)-1]))
        valuepart[strlen(valuepart)-1] = 0;
      
      definition[numDefinitions].categoryIndex = numCategories-1;
      strncpy(definition[numDefinitions].name, namepart, sizeof(definition[numDefinitions].name));
      strncpy(definition[numDefinitions].value, valuepart, sizeof(definition[numDefinitions].value));
      definition[numDefinitions].accessed = false;
      numDefinitions ++;
    }
  }

  if (infile != stdin)  
    fclose(infile);
}

bool ParameterFile::selectSection(const char *newCategory)
{
  for (int i=0; i<numCategories; i++) {
    if (!strcmp(categoryName[i], newCategory)) {
      selectedCategory = i;
      return true;
    }
  }
  
  return false;
}

bool ParameterFile::existsSetting(const char *name)
{
  for (int i=0; i<numDefinitions; i++) {
    if (definition[i].categoryIndex == selectedCategory) {
      if (!strcmp(name, definition[i].name))
        return true;
    }
  }
  
  return false;
}

void ParameterFile::getAsString(const char *name, char *buf, int bufsize)
{
  if (selectedCategory < 0)
    panic("No category selected");

  for (int i=0; i<numDefinitions; i++) {
    if (definition[i].categoryIndex == selectedCategory) {
      if (!strcmp(name, definition[i].name)) {
        strncpy(buf, definition[i].value, bufsize);
        definition[i].accessed = true;
        return;
      }
    }
  }
  
  panic("%s: Missing required setting '%s' in section [%s]", paramFileName, name, categoryName[selectedCategory]);
}

int ParameterFile::getAsInt(const char *name)
{
  char buf[20];
  getAsString(name, buf, sizeof(buf));
  return atoi(buf);
}

double ParameterFile::getAsDouble(const char *name)
{
  char buf[20];
  getAsString(name, buf, sizeof(buf));
  return atof(buf);
}

bool ParameterFile::getAsBoolean(const char *name)
{
  char buf[20];
  getAsString(name, buf, sizeof(buf));
  return (!strcasecmp(buf, "yes") || !strcasecmp(buf, "on") || !strcasecmp(buf, "true"));
}

void ParameterFile::assertAllAccessed()
{
  for (int i=0; i<numDefinitions; i++)
    if (!definition[i].accessed)
      panic("%s: Unknown setting '%s' in section [%s]", paramFileName, definition[i].name, categoryName[definition[i].categoryIndex]);
}

int ParameterFile::numSettingsInSection()
{
  assert(selectedCategory >= 0);
  int result = 0;
  for (int i=0; i<numDefinitions; i++)
    if (definition[i].categoryIndex == selectedCategory)
      result ++;
  return result;
}

void ParameterFile::getSettingByIndex(int index, char **name, char **value)
{
  assert((selectedCategory >= 0) && (index >= 0));
  for (int i=0; i<numDefinitions; i++) {
    if (definition[i].categoryIndex == selectedCategory) {
      if (index > 0) {
        index --;
      } else {
        *name = definition[i].name;
        *value = definition[i].value;
        definition[i].accessed = true;
        return;
      }
    }
  }

  *name = "";
  *value = "";
}

void ParameterFile::write(const char *filename)
{
  FILE *outfile;
  if (filename) {
    outfile = fopen(filename, "w+");
    if (!outfile)
      panic("Cannot write parameter file: '%s'", filename);
  } else {
    outfile = stdout;
  }

  int oldSelection = selectedCategory;
  bool needSpace = false;
  for (int i=0; i<numCategories; i++) {
    selectSection(categoryName[i]);
    if (numSettingsInSection() > 0) {
      if (needSpace)
        fprintf(outfile, "\n");
      else
        needSpace = true;
        
      fprintf(outfile, "[%s]\n", categoryName[i]);
      for (int j=0; j<numDefinitions; j++) {
        if (definition[j].categoryIndex == i)
          fprintf(outfile, "%s=%s\n", definition[j].name, definition[j].value);
      }
    }
  }
  
  selectedCategory = oldSelection;
  if (outfile != stdout)
    fclose(outfile);
}

void ParameterFile::set(const char *setCategory, const char *setName, const char *setValue)
{
  int icat = -1;
  for (int i=0; i<numCategories; i++)
    if (!strcmp(categoryName[i], setCategory))
      icat = i;

  if (icat == -1) {
    if (numCategories >= MAX_CATEGORIES)
      panic("Too many categories; cannot create new category '%s'", setCategory);
    strcpy(categoryName[numCategories], setCategory);
    icat = numCategories ++;
  }
  
  int idef = -1;
  for (int i=0; i<numDefinitions; i++) 
    if ((definition[i].categoryIndex == icat) && !strcmp(definition[i].name, setName))
      idef = i;
      
  if (idef == -1) {
    if (numDefinitions >= MAX_DEFINITIONS)
      panic("Too many definitions; cannot create new definition %s.%s", setCategory, setName);
    
    definition[numDefinitions].categoryIndex = icat;
    strcpy(definition[numDefinitions].name, setName);
    strcpy(definition[numDefinitions].value, setValue);
    idef = numDefinitions ++;
  }
  
  strcpy(definition[idef].value, setValue);
}
